<?php
/**
 * @package     CSVI
 * @subpackage  VirtueMart
 *
 * @author      RolandD Cyber Produksi <contact@rolandd.com>
 * @copyright   Copyright (C) 2006 - @year@ RolandD Cyber Produksi. All rights reserved.
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 * @link        https://rolandd.com
 */

namespace virtuemart\com_virtuemart\model\import;

defined('_JEXEC') or die;

/**
 * Category import.
 *
 * @package     CSVI
 * @subpackage  VirtueMart
 * @since       6.0
 */
class Category extends \RantaiImportEngine
{
	/**
	 * Category table
	 *
	 * @var    \VirtueMartTableCategory
	 * @since  6.0
	 */
	private $category = null;

	/**
	 * Media table
	 *
	 * @var    \VirtueMartTableMedia
	 * @since  6.0
	 */
	private $media = null;

	/**
	 * Category media table
	 *
	 * @var    \VirtueMartTableCategorymedia
	 * @since  6.0
	 */
	private $categoryMedia = null;

	/**
	 * Category language table
	 *
	 * @var    \VirtueMartTableCategoryLang
	 * @since  6.0
	 */
	private $categoryLang = null;

	/**
	 * Category model helper
	 *
	 * @var    \Com_virtuemartHelperCategory
	 * @since  6.0
	 */
	private $categoryModel = null;

	/**
	 * Category separator
	 *
	 * @var    string
	 * @since  6.0
	 */
	private $categorySeparator = null;

	/**
	 * Check if has_ fields exists
	 *
	 * @var    array
	 * @since  7.17.0
	 */
	private $hasFieldsExists = false;

	/**
	 * Run this before we start.
	 *
	 * @return  void.
	 *
	 * @since   7.17.0
	 * @throws  \Exception
	 */
	public function onBeforeStart()
	{
		// Check if there is has fields
		$this->hasFieldsExists = $this->helper->checkHasLinkedField('#__virtuemart_categories', 'has_children');
	}

	/**
	 * Here starts the processing.
	 *
	 * @return  bool  Akways returns true.
	 *
	 * @since   3.0
	 */
	public function getStart()
	{
		if (!$this->fields->getProcessRecord())
		{
			$this->loaded = false;

			return true;
		}

		$this->categoryModel = new \Com_virtuemartHelperCategory(
			$this->db,
			$this->template,
			$this->log,
			$this->csvihelper,
			$this->fields,
			$this->helper,
			$this->helperconfig,
			$this->userId
		);
		$this->categoryModel->getStart();

		// Check for vendor ID
		$this->setState('virtuemart_vendor_id', $this->helper->getVendorId());

		// Process data
		foreach ($this->fields->getData() as $fields)
		{
			foreach ($fields as $name => $details)
			{
				$value = $details->value;

				switch ($name)
				{
					case 'published':
					case 'media_published':
						switch ($value)
						{
							case 'n':
							case 'no':
							case 'N':
							case 'NO':
							case '0':
								$value = 0;
								break;
							default:
								$value = 1;
								break;
						}

						$this->setState($name, $value);
						break;
					case 'show_store_desc':
					case 'showcategory_desc':
					case 'showcategory':
					case 'showproducts':
					case 'omitLoaded':
					case 'showsearch':
					case 'featured':
					case 'omitLoaded_featured':
					case 'discontinued':
					case 'omitLoaded_discontinued':
					case 'latest':
					case 'omitLoaded_latest':
					case 'topten':
					case 'omitLoaded_topten':
					case 'recent':
					case 'omitLoaded_recent':
						switch (strtolower($value))
						{
							case 'n':
							case 'no':
							case '0':
								$value = 0;
								break;
							case 'y':
							case 'yes':
							case '1':
								$value = 1;
								break;
							default:
								$value = '';
								break;
						}

						$this->setState($name, $value);
						break;
					default:
						$this->setState($name, $value);
						break;
				}
			}
		}

		$this->loaded = true;

		// If we have no category path we cannot continue
		$categoryPath = $this->getState('category_path', false);

		if (!$categoryPath)
		{
			$this->log->addStats('incorrect', 'COM_CSVI_NO_CATEGORY_PATH_SET');

			$this->loaded = false;
		}

		return true;
	}

	/**
	 * Process a record.
	 *
	 * @return  bool  Returns true if all is OK | Returns false if no product SKU or product ID can be found.
	 *
	 * @since   6.0
	 */
	public function getProcessRecord()
	{
		if (!$this->loaded)
		{
			return true;
		}

		$translate = false;

		// Load the category separator
		if (is_null($this->categorySeparator))
		{
			$this->categorySeparator = $this->template->get('category_separator', '/');
		}

		// Set the category published state
		$this->categoryModel->category_publish = $this->getState('published', 1);

		// Loop through all categories if we are importing a translation
		if ($this->getState('category_path_trans', false))
		{
			$trans_paths = explode($this->categorySeparator, $this->getState('category_path_trans'));
			$paths       = explode($this->categorySeparator, $this->getState('category_path'));

			if (!is_array($paths))
			{
				$paths = (array) $paths;
			}

			$translate = true;
		}
		elseif ($this->template->get('language') == $this->template->get('target_language'))
		{
			$trans_paths = explode($this->categorySeparator, $this->getState('category_path'));
			$paths       = explode($this->categorySeparator, $this->getState('category_path'));
		}
		else
		{
			$this->log->addStats(
				'incorrect',
				\JText::sprintf(
					'COM_CSVI_CATEGORY_LANGUAGE_UNKNOWN',
					$this->template->get('language'),
					$this->template->get('target_language')
				)
			);

			return false;
		}

		// Get the count of categories in the path
		$countPaths = count($paths) - 1;

		// Get the last category
		$last = end($paths);
		reset($paths);
		$categoryDesc = $this->getState('category_description', '');

		// Process the paths
		foreach ($paths as $key => $path)
		{
			// Construct the full path
			$fullpath = array();

			for ($i = 0; $i <= $key; $i++)
			{
				$fullpath[] = $paths[$i];
			}

			$workPath = implode($this->categorySeparator, $fullpath);

			// First get the category ID
			try
			{
				if ($this->getState('virtuemart_category_id', 0))
				{
					$this->categoryModel->category_id = $this->getState('virtuemart_category_id', 0);
				}

				$categoryDetails = $this->categoryModel->getCategoryIdFromPath($workPath);
				$this->setState('virtuemart_category_id', $categoryDetails['category_id']);
				$categoryid = $this->getState('virtuemart_category_id', false);
			}
			catch (\Exception $e)
			{
				$this->log->addStats('incorrect', $e->getMessage());

				return false;
			}

			if ($categoryid && !$this->template->get('overwrite_existing_data'))
			{
				$this->log->add(\JText::sprintf('COM_CSVI_DATA_EXISTS_PRODUCT_SKU', $categoryid));
				$this->log->addStats('skipped', \JText::sprintf('COM_CSVI_DATA_EXISTS_PRODUCT_SKU', $categoryid));
			}
			elseif ($last === $path && $key === $countPaths)
			{
				// We have the category ID, lets see if it should be deleted
				if ($this->getState('category_delete', 'N') == 'Y')
				{
					$this->deleteCategory();
				}
				else
				{
					// Handle the images
					$this->processMedia();

					// Check on cat_params
					$this->processCategoryParams();

					// Set some basic values
					if (!$this->getState('modified_on', false))
					{
						$this->category->modified_on = $this->date->toSql();
						$this->category->modified_by = $this->userId;
					}

					// Check if the category_name matches the last entry in the category_path
					if ($this->getState('category_name', false))
					{
						if ($path != $this->getState('category_name'))
						{
							$this->log->addStats(
								'incorrect',
								\JText::sprintf(
									'COM_CSVI_CATEGORY_NAME_NO_MATCH_CATEGORY_PATH',
									$path,
									$this->getState('category_name')
								)
							);

							return false;
						}
					}

					// All fields have been processed, bind the data
					$this->category->bind($this->state);

					// Now store the data
					if (!$this->category->store())
					{
						$this->log->addStats(
							'incorrect',
							\JText::sprintf(
								'COM_CSVI_CATEGORY_DETAILS_NOT_ADDED',
								$this->category->getError()
							)
						);
					}

					// Store the language fields
					$this->categoryLang->load($this->category->virtuemart_category_id);
					$this->categoryLang->bind($this->state);

					// Set the translated category name
					if ($translate)
					{
						$this->categoryLang->set('category_name', $trans_paths[$key]);
						$this->setState('category_name', $trans_paths[$key]);
						$this->setState('category_description', $categoryDesc);
					}

					// Check and store the language data
					if ($this->categoryLang->check())
					{
						// Bind the new values
						$this->categoryLang->bind($this->state);

						// Recreate the slug
						if ($this->template->get('recreate_alias', false))
						{
							$this->categoryLang->createSlug();
						}

						if (!$this->categoryLang->store())
						{
							$this->log->addStats(
								'incorrect',
								\JText::sprintf(
									'COM_CSVI_CATEGORY_LANG_NOT_ADDED',
									$this->categoryLang->getError()
								)
							);

							return false;
						}
					}
					else
					{
						$this->log->addStats(
							'incorrect',
							\JText::sprintf(
								'COM_CSVI_CATEGORY_LANG_NOT_ADDED',
								$this->categoryLang->getError()
							)
						);

						return false;
					}
				}
			}
			else
			{
				// Do the parent category processing for transalation category path
				if ($categoryid && $translate)
				{
					// Category desc is needed only with last level
					$this->setState('category_description', '');

					// Store the language fields
					$this->categoryLang->load($categoryid);

					if($this->categoryLang->get('category_description'))
					{
						$this->setState('category_description', $this->categoryLang->get('category_description'));
					}

					$this->categoryLang->category_name = $trans_paths[$key];
					$this->categoryLang->bind($this->state);

					// Check and store the language data
					if ($this->categoryLang->check())
					{
						// Bind the new values
						$this->categoryLang->bind($this->state);

						// Recreate the slug
						if ($this->template->get('recreate_alias', false))
						{
							$this->categoryLang->createSlug();
						}

						if (!$this->categoryLang->store())
						{
							$this->log->addStats(
								'incorrect',
								\JText::sprintf(
									'COM_CSVI_CATEGORY_LANG_NOT_ADDED',
									$this->categoryLang->getError()
								)
							);

							return false;
						}
					}
					else
					{
						$this->log->addStats(
							'incorrect',
							\JText::sprintf(
								'COM_CSVI_CATEGORY_LANG_NOT_ADDED',
								$this->categoryLang->getError()
							)
						);

						return false;
					}
				}
			}

			// VM 3.8 update has_children field
			$this->updateParentHasValue($categoryid, 1);

			$this->category->reset();
			$this->categoryLang->reset();
		}
	}

	/**
	 * Load the necessary tables.
	 *
	 * @return  void.
	 *
	 * @since   6.0
	 *
	 * @throws  \CsviException
	 * @throws  \RuntimeException
	 */
	public function loadTables()
	{
		$this->category      = $this->getTable('Category');
		$this->media         = $this->getTable('Media');
		$this->categoryMedia = $this->getTable('CategoryMedia');

		// Check if the language tables exist
		$tables = $this->db->getTableList();

		// Get the language to use
		$language = $this->template->get('target_language');

		if ($this->template->get('language') === $this->template->get('target_language'))
		{
			$language = $this->template->get('language');
		}

		// Get the table name to check
		$tableName = $this->db->getPrefix() . 'virtuemart_categories_' . $language;

		if (!in_array($tableName, $tables))
		{
			$message = \JText::_('COM_CSVI_LANGUAGE_MISSING');

			if ($language)
			{
				$message = \JText::sprintf('COM_CSVI_TABLE_NOT_FOUND', $tableName);
			}

			throw new \CsviException($message, 510);
		}

		$this->categoryLang = $this->getTable('CategoryLang');
	}

	/**
	 * Clear the loaded tables.
	 *
	 * @return  void.
	 *
	 * @since   6.0
	 */
	public function clearTables()
	{
		$this->category->reset();
		$this->media->reset();
		$this->categoryMedia->reset();
		$this->categoryLang->reset();
	}

	/**
	 * Delete a category and its references.
	 *
	 * @return  void.
	 *
	 * @since   4.0
	 */
	private function deleteCategory()
	{
		// Update parent has_children field before deleting
		$this->updateParentHasValue($this->getState('virtuemart_category_id'), 0);

		// Delete the product
		if ($this->category->delete($this->getState('virtuemart_category_id')))
		{
			$this->log->addstats('deleted', 'com_csvi_category_deleted');

			// Delete category translations
			$languages = array_keys($this->csvihelper->getLanguages('lang_code'));

			foreach ($languages as $language)
			{
				$query = $this->db->getquery(true)
					->delete(
						$this->db->quotename('#__virtuemart_categories_' . strtolower(str_replace('-', '_', $language)))
					)
					->where(
						$this->db->quotename('virtuemart_category_id') . ' = ' . (int) $this->getState(
							'virtuemart_category_id'
						)
					);
				$this->db->setquery($query);
				$this->log->add('COM_CSVI_DEBUG_DELETE_CATEGORY_LANG_XREF', true);
				$this->db->execute();
			}

			// Delete category reference
			$query = $this->db->getquery(true)
				->delete($this->db->quotename('#__virtuemart_category_categories'))
				->where(
					$this->db->quotename('category_child_id') . ' = ' . (int) $this->getState('virtuemart_category_id')
				);
			$this->db->setquery($query);
			$this->log->add('COM_CSVI_DEBUG_DELETE_CATEGORY_XREF', true);
			$this->db->execute();

			// Delete media
			$query = $this->db->getquery(true)
				->delete($this->db->quotename('#__virtuemart_category_medias'))
				->where(
					$this->db->quotename('virtuemart_category_id') . ' = ' . (int) $this->getState(
						'virtuemart_category_id'
					)
				);
			$this->db->setquery($query);
			$this->log->add('COM_CSVI_DEBUG_DELETE_MEDIA_XREF', true);
			$this->db->execute();

			// Reset the products that link to this category
			$query = $this->db->getquery(true)
				->delete($this->db->quotename('#__virtuemart_product_categories'))
				->where(
					$this->db->quotename('virtuemart_category_id') . ' = ' . (int) $this->getState(
						'virtuemart_category_id'
					)
				);
			$this->db->setquery($query);
			$this->log->add('COM_CSVI_DEBUG_DELETE_PRODUCT_CATEGORY_XREF', true);
			$this->db->execute();

			// Reset any child categories to become parent
			$query = $this->db->getquery(true)
				->update($this->db->quotename('#__virtuemart_category_categories'))
				->set($this->db->quoteName('category_parent_id') . ' = 0')
				->where(
					$this->db->quotename('category_parent_id') . ' = ' . (int) $this->getState('virtuemart_category_id')
				);
			$this->db->setquery($query);
			$this->log->add('COM_CSVI_DEBUG_RESET_CATEGORY_PARENTS', true);
			$this->db->execute();
		}
		else
		{
			$this->log->addStats('incorrect', 'COM_CSVI_CATEGORY_NOT_DELETED');
		}
	}

	/**
	 * Process media files.
	 *
	 * @return  void.
	 *
	 * @since   4.0
	 */
	private function processMedia()
	{
		// Check if any image handling needs to be done
		if ($this->template->get('process_image', false))
		{
			if ($this->getState('file_url', false))
			{
				// Create an array of images to process
				$images       = explode('|', $this->getState('file_url'));
				$thumbs       = explode('|', $this->getState('file_url_thumb'));
				$titles       = explode('|', $this->getState('file_title'));
				$descriptions = explode('|', $this->getState('file_description'));
				$metas        = explode('|', $this->getState('file_meta'));
				$order        = explode('|', $this->getState('file_ordering'));
				$ordering     = 1;
				$max_width    = $this->template->get('resize_max_width', 1024);
				$max_height   = $this->template->get('resize_max_height', 768);

				// Image handling
				$imagehelper = new \CsviHelperImage($this->template, $this->log, $this->csvihelper);

				// Delete existing image links before importing new one
				if ($this->template->get('delete_category_images', false))
				{
					$query = $this->db->getQuery(true)
						->delete($this->db->quoteName('#__virtuemart_category_medias'))
						->where(
							$this->db->quoteName('virtuemart_category_id') . ' = ' . (int) $this->getState(
								'virtuemart_category_id'
							)
						);
					$this->db->setQuery($query)->execute();
					$this->log->add('Delete images for category');
				}

				foreach ($images as $key => $image)
				{
					$image = trim($image);

					$imgpath = $this->template->get(
						'file_location_category_images',
						'images/stories/virtuemart/category/'
					);

					// Make sure the final slash is present
					if (substr($imgpath, -1) != '/')
					{
						$imgpath .= '/';
					}

					if (!empty($image))
					{
						// Verify the original image
						if ($imagehelper->isRemote($image))
						{
							$original  = $image;
							$full_path = $imgpath;
						}
						else
						{
							// Check if the image contains the image path
							$dirname = dirname($image);

							if (strpos($imgpath, $dirname) !== false)
							{
								// Collect rest of folder path if it is more than image default path
								$imageleftpath = str_replace($imgpath, '', $dirname . '/');
								$image         = basename($image);

								if ($imageleftpath)
								{
									$image = $imageleftpath . $image;
								}
							}

							$original = $imgpath . $image;

							// Get subfolders
							$path_parts = pathinfo($original);
							$full_path  = $path_parts['dirname'] . '/';
						}

						// Generate image names
						if ($this->template->get('process_image', false))
						{
							$file_details = $imagehelper->ProcessImage($original, $full_path);
						}
						else
						{
							$file_details['exists']      = true;
							$file_details['isimage']     = $imagehelper->isImage(JPATH_SITE . '/' . $image);
							$file_details['name']        = $image;
							$file_details['output_name'] = basename($image);

							if (file_exists(JPATH_SITE . '/' . $image))
							{
								$file_details['mime_type'] = $imagehelper->findMimeType($image);
							}
							else
							{
								$file_details['mime_type'] = '';
							}

							$file_details['output_path'] = $full_path;
						}

						if ($file_details['exists'])
						{
							// Check if the image is an external image
							if (substr($file_details['name'], 0, 4) == 'http')
							{
								$this->log->addStats('incorrect', 'COM_CSVI_VM_NOSUPPORT_URL');
							}
							else
							{
								$title                         = (isset($titles[$key])) ? $titles[$key] : $file_details['output_name'];
								$description                   = (isset($descriptions[$key])) ? $descriptions[$key] : '';
								$meta                          = (isset($metas[$key])) ? $metas[$key] : '';
								$media                         = array();
								$media['virtuemart_vendor_id'] = $this->getState('virtuemart_vendor_id');

								if ($this->template->get('autofill'))
								{
									$media['file_title']       = $file_details['output_name'];
									$media['file_description'] = $file_details['output_name'];
									$media['file_meta']        = $file_details['output_name'];
								}
								else
								{
									$media['file_title']       = $title;
									$media['file_description'] = $description;
									$media['file_meta']        = $meta;
								}

								$media['file_mimetype']         = $file_details['mime_type'];
								$media['file_type']             = 'category';
								$media['file_is_product_image'] = 0;
								$media['file_is_downloadable']  = 0;
								$media['file_is_forSale']       = 0;
								$media['file_url']              = (empty($file_details['output_path'])) ? $file_details['output_name'] : $file_details['output_path'] . $file_details['output_name'];
								$media['published']             = $this->getState(
									'media_published',
									$this->getState('published', 0)
								);
								$media['file_is_product_image'] = 0;
								$media['file_url_thumb']        = '';

								// Create the thumbnail
								if ($file_details['isimage'])
								{
									$thumb = (isset($thumbs[$key])) ? $thumbs[$key] : null;

									if ($this->template->get('thumb_create'))
									{
										$thumb_sizes             = getimagesize(JPATH_SITE . '/' . $media['file_url']);
										$media['file_url_thumb'] = '';

										if ($thumb && ($thumb_sizes[0] < $max_width || $thumb_sizes[1] < $max_height))
										{
											$media['file_url_thumb'] = $imagehelper->createThumbnail(
												$media['file_url'],
												$this->template->get(
													'file_location_category_images'
												),
												$thumb
											);
										}
									}
									else
									{
										$resizeFolder = '';

										if (strpos($thumb, 'resized') !== false)
										{
											$resizeFolder = 'resized/';
										}

										$media['file_url_thumb'] = empty($thumb) ? $media['file_url'] : $file_details['output_path'] . $resizeFolder . basename(
												$thumb
											);

										if (substr($media['file_url_thumb'], 0, 4) == 'http')
										{
											$this->log->add(
												\JText::sprintf('COM_CSVI_RESET_THUMB_NOHTTP', $media['file_url_thumb'])
											);
											$media['file_url_thumb'] = '';
										}
									}
								}

								// Bind the media data
								$this->media->bind($media);

								// Check if the media image already exists
								$this->media->check();

								// Store the media data
								if ($this->media->store())
								{
									// Watermark the image
									if ($this->template->get('full_watermark') && $file_details['isimage'])
									{
										$imagehelper->addWatermark(JPATH_SITE . '/' . $media['file_url']);
									}

									// Store the category image relation
									$data                           = array();
									$data['virtuemart_category_id'] = $this->getState('virtuemart_category_id');
									$data['virtuemart_media_id']    = $this->media->get('virtuemart_media_id');
									$data['ordering']               = (isset($order[$key]) && !empty($order[$key])) ? $order[$key] : $ordering;
									$this->categoryMedia->bind($data);

									$this->categoryMedia->check();

									if ($this->categoryMedia->store())
									{
										$this->log->add('Store the category image relation', false);
										$ordering++;
									}

									// VM 3.8 update has_medias field
									if ($this->hasFieldsExists)
									{
										$this->setState('has_medias', 1);
									}
								}
								else
								{
									$this->log->addStats(
										'incorrect',
										\JText::sprintf(
											'COM_CSVI_MEDIA_NOT_ADDED',
											$this->media->getError()
										)
									);
								}

								// Reset the product media table
								$this->media->reset();
								$this->categoryMedia->reset();
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Process category params
	 *
	 * @return  void.
	 *
	 * @since   7.2.0
	 */
	private function processCategoryParams()
	{
		$newParam  = array();
		$catParams = array(
			'show_store_desc',
			'showcategory_desc',
			'showcategory',
			'categories_per_row',
			'showproducts',
			'omitLoaded',
			'showsearch',
			'productsublayout',
			'featured',
			'featured_rows',
			'omitLoaded_featured',
			'discontinued',
			'discontinued_rows',
			'omitLoaded_discontinued',
			'latest',
			'latest_rows',
			'omitLoaded_latest',
			'topten',
			'topten_rows',
			'omitLoaded_topten',
			'recent',
			'recent_rows',
			'omitLoaded_recent'
		);

		foreach ($catParams as $param)
		{
			$paramVal = $this->getState($param, '');

			if (isset($paramVal))
			{
				$newParam[] = $param . '="' . $this->getState($param, '') . '"';
			}
			else
			{
				$newParam[] = $param . '=""';
			}
		}

		$categoryParams = implode('|', $newParam);
		$this->setState('cat_params', $categoryParams);
	}

	/**
	 * Update parent category has value
	 *
	 * @param   int  $categoryId  Category id
	 * @param   int  $value       Value to update
	 *
	 * @return  void.
	 *
	 * @since   7.17.0
	 */
	private function updateParentHasValue($categoryId, $value)
	{
		$categoryXref = $this->getTable('CategoryXref');
		$categoryXref->load(array('category_child_id' => $categoryId));

		if ($categoryXref->category_parent_id > 0 && $this->hasFieldsExists)
		{
			$query = $this->db->getQuery(true)
				->update($this->db->quoteName('#__virtuemart_categories'))
				->set($this->db->quoteName('has_children') . ' = ' . (int) $value)
				->where(
					$this->db->quoteName('virtuemart_category_id') . ' = ' . (int) $categoryXref->category_parent_id
				);
			$this->db->setQuery($query)->execute();
			$this->log->add('Update has field for parent category');
		}
	}
}
